@@ -33,6 +33,14 @@ var inspectModuleCmd = &cobra.Command{
   timoni -n default inspect module app
 `,
 	RunE: runInspectModuleCmd,
+	ValidArgsFunction: func(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
+		switch len(args) {
+		case 0:
+			return completeInstanceList(cmd, args, toComplete)
+		default:
+			return nil, cobra.ShellCompDirectiveNoFileComp
+		}
+	},
 }
 
 type inspectModuleFlags struct {
